import os, litellm
from typing import List, Dict, Any

class LiteLLM():
    def __init__(self, model: str,):
        self.model = model
        self.base_url = os.environ.get("LITELLM_BASE_URL", "")
        self.api_key = os.environ.get("LITELLM_API_KEY", "")

    async def acompletion(self, messages: List=[], tools: List=[], stream: bool=False, **kwargs):
        metadata: Dict[str, Any] = {}
        custom_llm_provider = kwargs.get("custom_llm_provider", "openai")
        metadata["trace_id"] = kwargs.get("trace_id", None)
        parallel_tool_calls = kwargs.get("parallel_tool_calls", None)

        return await litellm.acompletion(
            model=self.model,
            messages=messages,
            tools=tools,
            stream=stream,
            parallel_tool_calls=parallel_tool_calls,
            api_base=self.base_url,
            api_key=self.api_key,
            custom_llm_provider=custom_llm_provider,
            metadata=metadata
        )

    def completion(self, messages: List=[], tools: List=[], **kwargs):
        metadata: Dict[str, Any] = {}
        custom_llm_provider = kwargs.get("custom_llm_provider", "openai")
        metadata["trace_id"] = kwargs.get("trace_id", None)
        parallel_tool_calls = kwargs.get("parallel_tool_calls", None)

        return litellm.completion(
            model=self.model,
            messages=messages,
            tools=tools,
            parallel_tool_calls=parallel_tool_calls,
            api_base=self.base_url,
            api_key=self.api_key,
            custom_llm_provider=custom_llm_provider,
            metadata=metadata
        )



from .agent import Agent
from .runner import Runner, RunnerStream
from .type import (
    RunResponse, StreamEvent, IterationData, ToolCallResponse
)
from .util import Util

__all__ = [
    "Agent",
    "Runner",
    "RunnerStream",
    "RunResponse",
    "StreamEvent",
    "IterationData",
    "ToolCallResponse",
    "Util"
]


from typing import Any, Dict, List, Optional, Callable
from .llms.provider import LITELLM

class Agent():
    def __init__(self,
        name: str,
        instructions: str,
        model: str,
        description: str = "",
        mcp_servers: Optional[Dict[str, Any]] = [],
        function_tools: Optional[List[Callable]] = [],
        sub_agents: Optional[List['Agent']] = [],
        parallel_tool_calls: Optional[bool] = None,
        llm_provider: Optional[str] = LITELLM,
        max_iterations: Optional[int] = 10,
        custom_llm_provider: Optional[str] = "openai",
        trace_id: Optional[str] = None
    ):
        self.name = name
        self.instructions = instructions
        self.model = model
        self.description = description
        self.mcp_servers = mcp_servers
        self.function_tools = function_tools
        self.sub_agents = sub_agents
        self.parallel_tool_calls = parallel_tool_calls
        self.llm_provider = llm_provider
        self.max_iterations = max_iterations
        self.custom_llm_provider = custom_llm_provider
        self.trace_id = trace_id

    def get_name(self) -> str:
        return self.name
    
    def get_instructions(self) -> str:
        return self.instructions
    
    def get_model(self) -> str:
        return self.model
    
    def get_description(self) -> str:
        return self.description
    
    def get_mcp_servers(self) -> Optional[Dict[str, Any]]:
        return self.mcp_servers
    
    def get_function_tools(self) -> Optional[List[Callable]]:
        return self.function_tools
    
    def get_sub_agents(self) -> Optional[List['Agent']]:
        return self.sub_agents
    
    def get_parallel_tool_calls(self) -> Optional[bool]:
        return self.parallel_tool_calls
    
    def get_llm_provider(self) -> Optional[str]:
        return self.llm_provider
    
    def get_max_iterations(self) -> Optional[int]:
        return self.max_iterations
    
    def get_custom_llm_provider(self) -> Optional[str]:
        return self.custom_llm_provider
    
    def get_trace_id(self) -> Optional[str]:
        return self.trace_id

class SubAgentManager():
    def __init__(self, agents: List[Agent]):
        self.agents = agents
        self.agent_name_map = {}
        self.tools = self.__load_agents_as_tools()
    
    def __load_agents_as_tools(self) -> List[Dict]:
        tools = []
        for agent in self.agents:
            toof_def = {
                "name": "sub_agent__" + agent.get_name(),
                "description": agent.get_description(),
                "parameters": {
                    "properties": {},
                    "required": [],
                    "type": "object"
                }
            }
            tools.append({
                "type": "function",
                "function": toof_def
            })
            self.agent_name_map[toof_def["name"]] = agent
        return tools
    
    def is_agent(self, name):
        if name in self.agent_name_map:
            return True
        return False

    def get_agents_as_tools(self) -> List[Dict]:
        return self.tools

    async def execute(self, runner_instance, agent_name, input: List[Dict[str, Any]]):
        agent = self.agent_name_map[agent_name]
        input = list(input) # list copy - to avoid deleting tool call from original run response result
        input.pop()
        return await runner_instance.run_sub_agent(agent, input)


import json, re

class Util:
    
    @classmethod
    def load_json(cls, json_string):
        # Cleaning Json String if there are back ticks or `json` keyword
        cleaned_string = re.sub(r'^```*|```$', '', json_string.strip(), flags=re.MULTILINE)
        cleaned_string = re.sub(r'^json*|$', '', cleaned_string.strip(), flags=re.MULTILINE)
        cleaned_string = cleaned_string.strip()
        
        try:
            # Converting cleaned JSON string to python `dict`
            return json.loads(cleaned_string)
        except json.JSONDecodeError as e:
            return f"Error parsing JSON: {e}"


import logging, json, asyncio, sys
from typing import List, Dict, Any
from .agent import Agent
from .model import Model
from .tool import Tool
from .type import (
    ModelSreamResponse, RunResponse, StreamEvent, ToolCallResponse, IterationData
)

class RunnerStream:

    ITER_START = "ITERATION_START"
    CONTENT_CHUNK = "CONTENT_CHUNK"
    TOOL_CALLS = "TOOL_CALLS"
    TOOL_RESPONSE = "TOOL_RESPONSE"
    ITER_END = "ITERATION_END"
    AGENT_RUN_END = "AGENT_RUN_END"
    MODEL_STREAM_COMPLETED = "MODEL_STREAM_COMPLETED"

    @classmethod
    def iteration_start(cls, data: int) -> StreamEvent:
        return StreamEvent(type=cls.ITER_START, data=data, data_type="int")
    
    @classmethod
    def content_chunk(cls, data: str) -> StreamEvent:
        return StreamEvent(type=cls.CONTENT_CHUNK, data=data, data_type="str")

    @classmethod
    def tool_calls(cls, data: List) -> StreamEvent:
        return StreamEvent(type=cls.TOOL_CALLS, data=data, data_type="List")

    @classmethod
    def tool_response(cls, data: ToolCallResponse) -> StreamEvent:
        return StreamEvent(type=cls.TOOL_RESPONSE, data=data, data_type="BaseModel")
    
    @classmethod
    def iteration_end(cls, data: IterationData) -> StreamEvent:
        return StreamEvent(type=cls.ITER_END, data=data, data_type="BaseModel")
    
    @classmethod
    def agent_run_end(cls, data: RunResponse) -> StreamEvent:
        return StreamEvent(type=cls.AGENT_RUN_END, data=data, data_type="BaseModel")
    
    @classmethod
    def model_stream_completed(cls, data: ModelSreamResponse) -> StreamEvent:
        return StreamEvent(type=cls.MODEL_STREAM_COMPLETED, data=data, data_type="BaseModel")
    
    @classmethod
    def data_dump(cls, event: StreamEvent) -> StreamEvent:
        if event.data_type == "int":
            return str(event.data)
        elif event.data_type == "str":
            return str(event.data)
        elif event.data_type == "List":
            return json.dumps(event.data)
        elif event.data_type == "Dict":
            return json.dumps(event.data)
        elif event.data_type == "BaseModel":
            return json.dumps(event.data.model_dump())

class Runner():
    def __init__(self,
        agent: Agent
    ):
        self.agent = agent
        self.mcp_manager = None
        self.ft_manager = None
        self.tool = None
        self.is_sub_agent = False
    
    def __set_agent_instructions(self, messages: List, system_prompt):
        if not system_prompt:
            return messages
        
        system_prompt_msg = {"role": "system", "content": system_prompt}
        if not messages or (len(messages) == 1 and messages[0].get("role") != "system"):
            messages.insert(0, system_prompt_msg)
        else:
            messages.append(system_prompt_msg)
        return messages

    async def __get_model_stream(self, response):
        model_stream_response = ModelSreamResponse(content="", tool_calls=[], message={"role": "assistant"})

        async for chunk in response:
            if hasattr(chunk, "choices") and len(chunk.choices) > 0:
                delta = chunk.choices[0].delta

                if hasattr(delta, "content") and delta.content:
                    model_stream_response.content += delta.content
                    yield RunnerStream.content_chunk(delta.content)

                if hasattr(delta, "tool_calls") and delta.tool_calls:
                    for tool_call in delta.tool_calls:
                        if tool_call.index >= len(model_stream_response.tool_calls):
                            model_stream_response.tool_calls.append({
                                "id": tool_call.id,
                                "type": "function",
                                "function": {
                                    "name": tool_call.function.name
                                    if hasattr(tool_call.function, "name")
                                    else "",
                                    "arguments": tool_call.function.arguments
                                    if hasattr(tool_call.function, "arguments")
                                    else "",
                                },
                            })
                        else:
                            if hasattr(tool_call.function, "arguments"):
                                model_stream_response.tool_calls[tool_call.index]["function"][
                                    "arguments"
                                ] += tool_call.function.arguments
                    
                    # Yield tool calls update
                    yield RunnerStream.tool_calls(model_stream_response.tool_calls)

        # Only add content if there is actual content
        if model_stream_response.content:
            model_stream_response.message["content"] = model_stream_response.content

        # Add tool calls if present
        if model_stream_response.tool_calls:
            model_stream_response.message["tool_calls"] = model_stream_response.tool_calls

        # Yield final complete response
        yield RunnerStream.model_stream_completed(model_stream_response)

    async def __stream_with_events(self, input: List[Dict[str, Any]]):
        """Internal streaming method that yields events"""
        input = self.__set_agent_instructions(input, self.agent.get_instructions())
        run_response = RunResponse(result=input, iterations=0)

        while run_response.iterations < self.agent.get_max_iterations():
            run_response.iterations += 1

            yield RunnerStream.iteration_start(run_response.iterations)

            response = await Model(self.agent.get_llm_provider()).run_async(
                model=self.agent.get_model(),
                messages=run_response.result,
                tools=self.tool.get_tools(),
                stream=True,
                parallel_tool_calls = self.agent.get_parallel_tool_calls(),
                custom_llm_provider=self.agent.get_custom_llm_provider(),
                trace_id=self.agent.get_trace_id()
            )

            # Consume the stream and emit events for each chunk
            stream_response: ModelSreamResponse = None
            async for stream_event in self.__get_model_stream(response):
                if stream_event.type == RunnerStream.MODEL_STREAM_COMPLETED:
                    stream_response = stream_event.data
                else:
                    yield stream_event

            # Append the complete message to result
            run_response.result.append(stream_response.message)

            iteration_data = IterationData(
                iterations=run_response.iterations,
                has_tool_calls=bool(stream_response.tool_calls)
            )

            logging.info(
                f"Iteration {run_response.iterations}: Received response - "
                f"content length: {len(stream_response.content)} chars, tool_calls: {len(stream_response.tool_calls)}"
            )

            if not stream_response.tool_calls:
                logging.info(f"No tool calls made. Agent finished after {run_response.iterations} iterations.")
                # Yield agent finished event
                yield RunnerStream.iteration_end(iteration_data)
                await self.tool.close_mcp_manager()
                run_response.sub_agents_response = self.tool.get_sub_agents_response()
                yield RunnerStream.agent_run_end(run_response)
                return

            tool_responses: List[ToolCallResponse] = await self.tool.tool_calls(
                stream_response.tool_calls, run_response.result
            )

            for tool_response in tool_responses:
                run_response.result.append(tool_response.model_dump())
                # Yield tool response event
                yield RunnerStream.tool_response(tool_response)

            # Yield iteration end event
            yield RunnerStream.iteration_end(iteration_data)

        # Yield agent finished event if max iterations reached
        await self.tool.close_mcp_manager()
        run_response.sub_agents_response = self.tool.get_sub_agents_response()
        run_response.max_iterations_reached = True
        yield RunnerStream.agent_run_end(run_response)

    async def run_stream(self, input: List[Dict[str, Any]] = [{}]):
        try:
            self.tool = await Tool(self).init_tools(self.agent)
            async for event in self.__stream_with_events(input):
                yield event
        except Exception as e:
            if self.tool:
                await self.tool.close_mcp_manager()
            raise

    def __parse_model_response(self, response) -> Dict:
        final_response = {"role": "assistant", "content": "", "tool_calls": []}

        if hasattr(response, "choices") and len(response.choices) > 0:
            res = response.choices[0].message

            if hasattr(res, "content") and res.content:
                final_response["content"] += res.content

            if hasattr(res, "tool_calls") and res.tool_calls:
                for tool_call in res.tool_calls:
                    final_response["tool_calls"].append({
                        "id": tool_call.id,
                        "type": "function",
                        "function": {
                            "name": tool_call.function.name
                            if hasattr(tool_call.function, "name")
                            else "",
                            "arguments": tool_call.function.arguments
                            if hasattr(tool_call.function, "arguments")
                            else "",
                        },
                    })
        if not final_response["content"]:
            del final_response["content"]
        
        if not final_response["tool_calls"]:
            del final_response["tool_calls"]

        return final_response

    async def __execute(self, input: List[Dict[str, Any]]):
        input = self.__set_agent_instructions(input, self.agent.get_instructions())
        run_response = RunResponse(result=input, iterations=0)
        # If sub agent, get the last index value of the input. It will be the system prompt in any way.
        if self.is_sub_agent and self.agent.get_instructions():
            run_response.sub_agent_result.append(input[-1])

        while run_response.iterations < self.agent.get_max_iterations():
            run_response.iterations += 1
            response = Model(self.agent.get_llm_provider()).run(
                model=self.agent.get_model(),
                messages=run_response.result,
                tools=self.tool.get_tools(),
                parallel_tool_calls = self.agent.get_parallel_tool_calls(),
                custom_llm_provider=self.agent.get_custom_llm_provider(),
                trace_id=self.agent.get_trace_id()
            )

            response = self.__parse_model_response(response)
            
            run_response.result.append(response)

            if self.is_sub_agent:
                run_response.sub_agent_result.append(response)

            if not response.get("tool_calls", []):
                run_response.sub_agents_response = self.tool.get_sub_agents_response()
                return run_response

            tool_responses: List[ToolCallResponse] = await self.tool.tool_calls(
                response.get("tool_calls"), run_response.result
            )

            for tool_response in tool_responses:
                run_response.result.append(tool_response.model_dump())

        run_response.sub_agents_response = self.tool.get_sub_agents_response()
        run_response.max_iterations_reached = True
        return run_response

    async def run_async(self, input: List[Dict[str, Any]]):
        try:
            # If caller function is 'run_sub_agent', its a sub agent call
            if (sys._getframe(1).f_code.co_name) == 'run_sub_agent':
                self.is_sub_agent = True
            self.tool = await Tool(self).init_tools(self.agent)
            exec_result = await self.__execute(input)
            await self.tool.close_mcp_manager()
            return exec_result
        except Exception as e:
            if self.tool:
                await self.tool.close_mcp_manager()
            raise

    def run(self, input: List[Dict[str, Any]] = [{}]):
        try:
            return asyncio.run(self.run_async(input))
        except Exception as e:
            raise

    @classmethod
    async def run_sub_agent(cls, agent: Agent, input=[{}]):
        return await cls(agent).run_async(input=input)



from typing import List, Dict, Any
from pydantic import BaseModel


class ModelSreamResponse(BaseModel):
    content: str
    tool_calls: List
    message: Dict[str, Any]

class RunResponse(BaseModel):
    result: List[Dict[str, Any]]
    iterations: int
    sub_agent_result: List[Dict[str, Any]] = []
    sub_agents_response: Dict[str, Any] = {}
    max_iterations_reached: bool = False

class IterationData(BaseModel):
    iterations: int
    has_tool_calls: bool

class StreamEvent(BaseModel):
    type: str
    data: Any
    data_type: str = "none"

class ToolCallResponse(BaseModel):
    role: str
    tool_call_id: str
    content: Any


from stark import Agent, Runner

def main():

    agent_3 = Agent(
        name="Delivery-Agent",
        description="This agent will be used for delivering pizza", # Sub agent must have a description.
        instructions="Say `pizza has been delivered` in your response and add pizza type and temperature if available",
        model="gemini-3-pro-preview"
    )

    
    agent_2 = Agent(
        name="Pizza-Agent",
        description="This agent will be used for baking pizza", # Sub agent must have a description.
        instructions="Say 'pizza has been baked at 60 degree' in your response, add pizza type in the response if available and call Delivery Agent",
        model="gemini-3-pro-preview"
    )

    agent_1 = Agent(
        name="Master-Agent",
        instructions="You have sub_agent tool. Call tools based on user query",
        model="gemini-3-pro-preview",
        sub_agents=[agent_2, agent_3]
    )

    result = Runner(agent_1).run(input=[{ "role": "user", "content": "I need to order peperoni pizza" }])

    print(result)
    print("")
    print(result.sub_agents_response.get("Pizza-Agent"))
    print("")
    print(result.sub_agents_response.get("Delivery-Agent"))

main()


import shutil
from typing import Dict, Any, List
from contextlib import AsyncExitStack
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from mcp.types import Tool

class StdioMCP():
    """
    Manages multiple persistent MCP server connections.
    """
    def __init__(self):
        self.sessions: ClientSession = None
        self._exit_stack = AsyncExitStack()

    def __format_tools_for_input(self, tools: List[Tool]) -> List:
        tools_output: List[Any] = []
        for tool in tools:
            tools_output.append({
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description or "",
                    "parameters": tool.inputSchema
                    if hasattr(tool, "inputSchema")
                    else {"type": "object", "properties": {}},
                },
            })
        return tools_output

    async def connect_server(self, name:str, config: Dict[str, Any], _exit_stack: AsyncExitStack):
        """
        Spawns multiple servers and keeps their sessions active.
        
        Args:
            server_configs: Dict where key is server_name and value is a dict
                            containing 'command', 'args', and optional 'env'.
        """
        command = config.get("command")
        args = config.get("args", [])
        env = config.get("env", None)

        # Check if command exists (e.g., 'python', 'npx')
        if not shutil.which(command):
            print(f"⚠️  Warning: Command '{command}' not found. Skipping {name}.")
            return None

        print(f"🔌 Connecting to {name} via {command}...")

        server_params = StdioServerParameters(
            command=command,
            args=args,
            env=env
        )

        try:
            stdio_transport = await _exit_stack.enter_async_context(
                stdio_client(server_params)
            )
            read, write = stdio_transport

            # 2. Create the session
            session = await _exit_stack.enter_async_context(
                ClientSession(read, write)
            )
                
            # 3. Initialize the protocol
            await session.initialize()

            # 4. List tools
            self.tools = self.__format_tools_for_input((await session.list_tools()).tools)
            self.session = session

            print(f"✅ {name} connected and initialized.")
            return _exit_stack

        except Exception as e:
            print(f"❌ Failed to connect to {name}: {e}")
            # Clean up on error
            raise

    async def call_tool(self, tool_name: str, arguments: dict = None):
        """
        Calls a tool on a specific preserved session.
        """
        result = await self.session.call_tool(tool_name, arguments or {})
        return result

    def get_tools(self):
        """Helper to see what tools are available on a server."""
        return self.tools

    def get_session(self):
        """Helper to see what tools are available on a server."""
        return self.session



import os, json
from stark.agent import Agent
from stark.runner import Runner
from stark.llms import LITELLM

MCP_SERVER = {
    "atlassian": {
        "command": "uvx",
        "args": ["mcp-atlassian"],
        "env": {
            "JIRA_URL": os.environ.get("JIRA_URL", ""),
            "JIRA_USERNAME": os.environ.get("JIRA_EMAIL", ""),
            "JIRA_API_TOKEN": os.environ.get("JIRA_TOKEN", "")
        }
    },
    "slack": {
        "command": "uvx",
        "args": ["mcp-slack"],
        "env": {
            "SLACK_BOT_TOKEN": os.environ.get("SLACK_BOT_TOKEN", "")
        }
    }
}

def stark_search(input: str):
    """
    {
        "description": "This function will search the stark logs",
        "parameters": {
            "properties": {
                "query": {
                    "description": "Query send by user to search in stark logs",
                    "title": "User Query",
                    "type": "string"
                }
            },
            "required": [
                "query"
            ],
            "type": "object"
        }
    }
    """
    try:
        if isinstance(input, str):
            input = json.loads(input)
        if "query" in input:
            return json.dumps({"result": "This is a mock function with Query"})
        return json.dumps({"result": "This is a mock function without Query"})
    except Exception as e:
        return f"Exception in calling the function: {str(e)}"

def main():
    user_query = "Stark AI is not working"
    ticket_classification = "stark_issue"
    slack_channel = "#random"
    instructions = f"""### Context
You are an expert Assistant. Your role is to manage Jira Tickets and search stark logs.

### Ask
1. Create a Jira issue with following fields:
    1.1. Project Key: STARK
    1.2. Issue Type: 'Task'
    1.3. Summary: {user_query}
    1.4. Description:
        ```
        __Description:__ {user_query}
        ```
    1.5. Labels: type::{ticket_classification}
2. Transition issue to 'In-Progress' status
3. Send ticket `issue_key` to slack channel: `{slack_channel}`
4. Search stark logs using following user query: {user_query}

### Rules
1. Don't output anything else other than the JSON string.
    1.1. Correct Output Example: {{"issue_key": "STARK-12345"}}
    1.2. Incorrect Output Example:  ```json\\n{{\\n"issue_key": "STARK-12345"\\n}}\\n```
2. Return following key as JSON output:
    2.1. `issue_key`
3. Pass `issue_key` to slack message and replace place holder `{{ issue_key }}` with it.
"""

    try:
        agent = Agent(
            name="Jira-Support",
            instructions=instructions,
            model="claude-sonnet-4-5",
            mcp_servers=MCP_SERVER,
            function_tools=[stark_search],
            llm_provider=LITELLM
        )

        result = Runner(agent).run(input=[{ "role": "user", "content": user_query }])
        
        print(result)

    except Exception as e:
        raise

main()



from typing import List
from .llms.provider import LITELLM
from .llms.litellm import LiteLLM


class Model():
    def __init__(self, provider):
        self.provider = provider

    async def run_async(
        self, model: str,
        messages: List=[],
        tools: List=[],
        stream: bool=False,
        **kwargs
    ):
        if self.provider == LITELLM:
            return  await LiteLLM(model).acompletion(
                messages=messages,
                tools=tools,
                stream=stream,
                **kwargs
            )
    
    def run(
        self, model: str,
        messages: List=[],
        tools: List=[],
        **kwargs
    ):
        if self.provider == LITELLM:
            return LiteLLM(model).completion(
                messages=messages,
                tools=tools,
                **kwargs
            )


import logging, json
from typing import List, Dict, Any
from .mcp import MCPManager
from .function import FunctionToolManager
from .agent import Agent, SubAgentManager
from .type import ToolCallResponse, RunResponse

class Tool:
    def __init__(self, runner):
        self.runner = runner
        self.mcp_manager = None
        self.ft_manager = None
        self.sub_agent_manager = None
        self.tools = []
        self.sub_agents_response = {}

    async def init_tools(self, agent: Agent):
        mcp_servers = agent.get_mcp_servers()
        function_tools = agent.get_function_tools()
        sub_agents = agent.get_sub_agents()
        if mcp_servers:
            self.mcp_manager = await MCPManager.init(mcp_servers)
            self.tools = self.tools + self.mcp_manager.get_tools()
        if function_tools:
            self.ft_manager = FunctionToolManager(function_tools)
            self.tools = self.tools + self.ft_manager.get_tools()
        if sub_agents:
            self.sub_agent_manager = SubAgentManager(sub_agents)
            self.tools = self.tools + self.sub_agent_manager.get_agents_as_tools()
        return self

    def get_tools(self) -> List[Dict]:
        return self.tools

    async def close_mcp_manager(self):
        if self.mcp_manager:
            await self.mcp_manager.close_all_sessions()
    
    def get_sub_agents_response(self) -> Dict:
        return self.sub_agents_response
    
    async def tool_calls(
        self, ai_tool_calls,
        messages: List[Dict[str, Any]] = [{}]
    ) -> List[ToolCallResponse]:
        tool_responses: List[ToolCallResponse] = []
        for ai_tool_call in ai_tool_calls:
            tool_responses.append(await self.__call(ai_tool_call, messages))
        return tool_responses

    async def __call(
        self, ai_tool_call: Dict,
        messages: List[Dict[str, Any]] = [{}]
    ) -> ToolCallResponse:
        tool_name: str = ai_tool_call["function"]["name"]
        tool_call_id = ai_tool_call["id"]
        tool_result = None

        try:
            arguments = json.loads(ai_tool_call["function"]["arguments"])
        except json.JSONDecodeError as e:
            logging.error(f"Failed to parse arguments for {tool_name}: {e}")
            arguments = {}

        if (self.mcp_manager and self.mcp_manager.is_mcp_tool(tool_name)) and (self.ft_manager and self.ft_manager.is_function_tool(tool_name)):
            tool_result = f"Tool name ({tool_name}) didn't execute because same tool exist in one of the MCP servers and in one of the function tools"
            return ToolCallResponse(role="tool", tool_call_id=tool_call_id, content=tool_result)

        logging.info(f"🔧 Tool request: {tool_call_id}")
        logging.info(f"🔧 Tool name: {tool_name}")
        logging.info(f"🔧 Tool Args: {arguments}")

        if self.mcp_manager and self.mcp_manager.is_mcp_tool(tool_name):
            try:
                tool_error = None
                result = await self.mcp_manager.call_tool(tool_name, arguments)

                if result.content:
                    if hasattr(result.content[0], "text"):
                        tool_result = result.content[0].text
                    elif hasattr(result.content[0], "data"):
                        tool_result = str(result.content[0].data)
                    else:
                        tool_result = str(result.content[0])
                else:
                    tool_result = ""
                
                # Check if result is an error from wrong server
                if tool_result and "Unknown tool:" in tool_result:
                    logging.warning(
                        f"Tool {tool_name} not available"
                        f"trying next server"
                    )

                logging.info(f"Tool {tool_name} result length: {len(tool_result)} chars")
            except Exception as e:
                tool_error = str(e)
                logging.info(
                    f"Tool {tool_name} raised exception: {e}"
                )

            # Ensure tool_result is never empty
            if tool_result is None or (isinstance(tool_result, str) and not tool_result.strip()):
                if tool_result is None:
                    error_msg = (
                        f"Tool {tool_name} not found"
                        if not tool_error
                        else f"Tool error: {tool_error}"
                    )
                    logging.error(f"Failed to execute {tool_name}: {error_msg}")
                    tool_result = json.dumps({"error": error_msg})
                else:
                    # Empty result - provide a default message
                    tool_result = "Tool executed successfully (no output returned)"
                    logging.warning(f"Tool {tool_name} returned empty result")

            tool_result = tool_result if isinstance(tool_result, str) else json.dumps(tool_result)
            if not tool_result.strip():
                tool_result = "Tool executed successfully (no output returned)"

        if self.ft_manager and self.ft_manager.is_function_tool(tool_name):
            tool_result = self.ft_manager.call_tool(tool_name, arguments)
            if not isinstance(tool_result, str):
                tool_result = str(tool_result)

        if self.sub_agent_manager and self.sub_agent_manager.is_agent(tool_name):
            tool_result: RunResponse = await self.sub_agent_manager.execute(self.runner, tool_name, messages)
            tool_result = tool_result.sub_agent_result
            self.sub_agents_response.update({tool_name.removeprefix("sub_agent__"): tool_result[-1]})
            logging.info(tool_result)
            if not isinstance(tool_result, str):
                tool_result = str(tool_result)

        return ToolCallResponse(role="tool", tool_call_id=tool_call_id, content=tool_result)



from typing import Dict, Any, List
from contextlib import AsyncExitStack
from .mcp_servers.stdio import StdioMCP

class MCPManager:
    """
    Manages multiple persistent MCP server connections.
    """

    @classmethod
    async def init(cls, mcp_servers: Dict[str, Dict[str, Any]]):
        if mcp_servers:
            instance = cls(mcp_servers)
            await instance.connect_servers()
            return instance
        return None

    def __init__(self, server_configs: Dict[str, Dict[str, Any]]):
        self.servers: Dict[str, Any |  StdioMCP] = {}
        if server_configs:
            self.server_configs = server_configs
            self.tool_to_server: Dict[str, str]= {}
            self._exit_stack = AsyncExitStack()
    
    def __map_tool_name_to_server_name(self, server_name: str, tools: List[Dict]):
        for tool in tools:
            self.tool_to_server[tool.get("function").get("name")] = server_name

    async def connect_servers(self):
        for name, config in self.server_configs.items():
            transport = config.get("transport", "stdio")
            server = StdioMCP()
            if transport == "streamable_http":
                server = None
            await server.connect_server(name, config, self._exit_stack)
            self.servers[name] = server
            self.__map_tool_name_to_server_name(name, server.get_tools())
    
    async def call_tool(self, tool_name: str, arguments: dict = None):
        if self.servers:
            return await self.servers[self.tool_to_server[tool_name]].call_tool(tool_name, arguments)
        return None

    def get_tools(self) -> List[Dict]:
        if self.servers:
            return [tool for _, server in self.servers.items() for tool in server.get_tools()]
        return []
    
    async def close_all_sessions(self):
        """Close all MCP server sessions properly."""
        await self._exit_stack.aclose()

    def is_mcp_tool(self, tool_name: str):
        if tool_name in self.tool_to_server:
            return True
        return False


import json
from typing import List, Callable, Dict, Any

class FunctionToolManager:
    def __init__(self, function_tools: List[Callable]):
        self.function_tools = function_tools
        self.func_name_map = {}
        self.tools = self.__load_tools()

    def __is_valid_json(self, json_str):
        try:
            json.loads(json_str)
            return True
        except json.decoder.JSONDecodeError:
            return False

    def __load_tools(self) -> List[Dict]:
        tools = []
        for function_tool in self.function_tools:
            if callable(function_tool) and self.__is_valid_json(function_tool.__doc__):
                toof_def = json.loads(function_tool.__doc__)
                toof_def["name"] = function_tool.__name__
                tools.append({
                    "type": "function",
                    "function": toof_def
                })
                self.func_name_map[function_tool.__name__] = function_tool
        return tools
    
    def get_tools(self) -> List[Dict]:
        return self.tools

    def call_tool(self, tool_name: str, arguments: Any):
        func = self.func_name_map[tool_name]
        return func(arguments)
    
    def is_function_tool(self, tool_name):
        if tool_name in self.func_name_map:
            return True
        return False

